% Created 2010-May-05
% compute the BFGS direction for the structured H2 problem
% input: difference of two gradient steps;
%        difference of two steps;
%        approximation of hessian inverse;
%        information structure identity;
%        the gradient of current step.
% the code updates the inverse of Hessian by BFGS formula 
% the advantage is to avoid the inversion of Hessian
% See Nocedal and Wright page 197 (8.16)

% 

function [Fbfgs,Hs] = SH2_BFGSdir(GradDiff,Fdiff,Hs,IS,gradF)

% get the size of the feedback gain

    [n,m] = size(Fdiff);

% vectorize the structural identity

    vecIS = vec(IS);

% find the indexes of 1's in the vector form of IS

    idxIS = find(vecIS==1);

% select the structured entries and put them in vectors

    s = Fdiff(idxIS);       
    y = GradDiff(idxIS);  
    ro = 1/(y'*s);
    gF = gradF(idxIS);      
    
% check the curvature condition, because the problem may not be convex

    if ro < 0 
        disp('curvature condition is not satisfied')
    end
    
% compute the approximation of the inverse Hessian    

    Hs = Hs - ro*s*y'*Hs - Hs*ro*y*s' + (ro^2*y'*Hs*y+ro)*s*s';

% experiments about the computation efficiency of different schemes
% apparently, scheme (2) is the most efficient from both complexity
% analysis and experiment results.

% % (1)BFGS: two matrix multiplication here O(n^3)
% tic
% Hs = (I-ro*s*y')*Hs*(I-ro*y*s') + ro*s*s';
% bfgs = -Hs*gF;
% toc
% 
% % % (2)BFGS: matrix-vector multiplications and matrix sum here O(n^2)
% tic
% Hs = Hs - ro*s*y'*Hs - Hs*ro*y*s' + (ro^2*y'*Hs*y+ro)*s*s';
% bfgs = -Hs*gF;
% toc
% 
% % % (3)DFP: one matrix inverse here O(n^3)
% tic
% Hs = Hs + y*y'/(y'*s) - Hs*s*s'*Hs/(s'*Hs*s);
% bfgs = -Hs\gF;
% toc

% minEigHs = min(eig(Hs));
% if minEigHs<=0
%     disp('inverse Hessian is not positive definite!')
% end

% compute the BFGS direction and convert it into the sparse matrix form

    bfgs = -Hs*gF;

    vecbfgs = zeros(n*m,1);

    for ind=1:length(idxIS)    
            vecbfgs(idxIS(ind)) = bfgs(ind);
    end

    Fbfgs = reshape(vecbfgs,n,m);