% created 2010-May-05
% the code solves the structured H2 problem using gradient method
% you need to set up the problem;
% input required for the code:
% A, B1, B2, C2, Q, R and structural identity IS.
% the output is the optimal objective function Jopt and the optimal
% structured gain Kopt.

% backtrapping line search 

    alpha = 0.3;
    beta = 0.5;

% tolerance of norm of gradient direction
% it is difficult for gradient method to achieve high accurate solution,
% i.e., tol should not be too small

    tol = 1.e-5;

% number of gradient steps    
    
    GradStep = 0;

% the complementary structural identity
    
    ISc = ones(size(IS))-IS;

% centralized controller

    [Kc,S,E] = lqr(A,B2,Q,R);


% check the number of output 

[pn,nn] = size(C2);

% case 1, if C2 is skinny full column-rank, initialize the descent method with 
% K = Kc*inv(C2'*C2)*C2';

if pn >= nn
    
     K0 = (Kc*inv(C2'*C2)*C2').*IS; 
     
end

% case 2, if C2 is fat full row-rank, do the coordinate transformation

if pn < nn
    
    T = [C2; null(C2)'];
    invT = inv(T);
    A = T*A*invT;  
    B1=T*B1; B2=T*B2;
    Q = invT'*Q*invT;  
    [Kc,S,E] = lqr(A,B2,Q,R);

    % after the change of coordiantes, C2 = [I O];
    % then C2 should be eliminated, but to make the code
    % consistent we have C2=I
    % the initial condition is the truncated optimal feedback gain Kc.*IS
    
    C2 = eye(size(A));
    IS = [IS, zeros(pn,nn-pn)];    
    ISc = ones(size(IS)) - IS;
    K0 = Kc.*IS;
    
end
       

% start timer

tic

    
    disp('GradStep      ngradK           J')       

% compute the H2 norm for the initial gain
    
    J0 = trace(lyap((A - B2*K0*C2)',Q + C2'*K0'*R*K0*C2)*B1*B1');
    
    K = K0;
    J = J0;
    
% compute the gradient 

    P = lyap((A - B2*K*C2)',Q + C2'*K'*R*K*C2); 
    L = lyap((A - B2*K*C2),B1*B1'); 
    gradK = 2*(R*K*C2*L*C2' - B2'*P*L*C2').*IS;
    ngradK = norm(gradK,'fro');
    GradStep = GradStep+1;
%     
    
    stepsize = 1;
    Ktemp = K - stepsize*gradK;

    maxEigAcl = max(real(eig(A - B2*Ktemp*C2)));
    Jtemp = trace(lyap((A - B2*Ktemp*C2)',Q + C2'*Ktemp'*R*Ktemp*C2)*B1*B1');

% use backtracking line search and also check for closed-loop stability    
    
    while 1

        while maxEigAcl >= 0 || J - Jtemp < stepsize*alpha*trace(gradK'*gradK)
            stepsize = beta*stepsize;
            if stepsize<1.e-19
                if max(real(eig(A - B2*(Ktemp.*IS)*C2)))<0
                    Kp=Ktemp.*IS;    %#ok<NASGU>
                    Jp=trace(lyap((A - B2*Kp*C2)',Q + C2'*Kp'*R*Kp*C2)*B1*B1');        %#ok<NASGU>
                end         
                toc
                error('Gradient method is stuck with very small stepsize! Try to increase tol')
            end
            Ktemp = K - stepsize*gradK;
            maxEigAcl = max(real(eig(A-B2*Ktemp*C2)));
            Jtemp = trace(lyap((A-B2*Ktemp*C2)', Q + C2'*Ktemp'*R*Ktemp*C2)*B1*B1');
        end
        
% update the current step K
        
        K = Ktemp;
        J = Jtemp;
        
% compute the gradient of current step

        P = lyap((A-B2*K*C2)',Q + C2'*K'*R*K*C2);   
        L = lyap((A-B2*K*C2),B1*B1');    

        gradK = 2*(R*K*C2*L*C2' - B2'*P*L*C2').*IS;        
        ngradK = norm(gradK,'fro');
        GradStep = GradStep+1;

% display the number of gradient steps, the norm of gradient, and the
% current objective function value J
        
        disp([num2str(GradStep),'         ',num2str(ngradK,'%6.4E'),'     ',num2str(J,'%6.4E')])

% stop the algorithm if the norm of gradient is small enough        
        
        if ngradK<tol
            break;
        end
        
% otherwise, compute the next gradient step

        stepsize = 1;
        Ktemp = K - stepsize*gradK;
        maxEigAcl = max(real(eig(A-B2*Ktemp*C2)));
        Jtemp = trace(lyap((A-B2*Ktemp*C2)',Q + C2'*Ktemp'*R*Ktemp*C2)*B1*B1');

        
    end

    
% end timer

toc


% if the output matrix C2 has full-row rank, K = [K1 K2] with
% K1.*IS = K1 and K2=0;

if pn<nn
    Kopt = K(1:pn,1:pn);
else
    Kopt = K;
end

Jopt = J;